@@ -1,4 +1,4 @@
-# Copyright 2022 The dataclass_array Authors.
+# Copyright 2023 The dataclass_array Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -25,7 +25,7 @@ from dataclass_array import field_utils
 from dataclass_array import shape_parsing
 from dataclass_array import type_parsing
 from dataclass_array.typing import Array
-from dataclass_array.typing import Axes, DcOrArray, DcOrArrayT, DTypeArg, DynamicShape, Shape  # pylint: disable=g-multiple-import
+from dataclass_array.typing import Axes, DTypeArg, DcOrArray, DcOrArrayT, DynamicShape, Shape  # pylint: disable=g-multiple-import,g-importing-member
 from dataclass_array.utils import np_utils
 from dataclass_array.utils import py_utils
 import einops
@@ -37,6 +37,7 @@ import numpy as np
 import typing_extensions
 from typing_extensions import Annotated, Literal, TypeAlias  # pylint: disable=g-multiple-import
 
+
 lazy = enp.lazy
 
 # TODO(pytype): Should use `dca.typing.DcT` but bound does not work across
@@ -83,7 +84,6 @@ def dataclass_array(
 
   ```python
   @dca.dataclass_array()
-  @dataclasses.dataclass(frozen=True)
   class MyDataclass(dca.DataclassArray):
     ...
   ```
@@ -116,6 +116,31 @@ def dataclass_array(
   return decorator
 
 
+def array_field(
+    shape: Shape,
+    dtype: DTypeArg = float,
+    **field_kwargs,
+) -> dataclasses.Field[DcOrArray]:
+  """Dataclass array field.
+
+  See `dca.DataclassArray` for example.
+
+  Args:
+    shape: Inner shape of the field
+    dtype: Type of the field
+    **field_kwargs: Args forwarded to `dataclasses.field`
+
+  Returns:
+    The dataclass field.
+  """
+  # TODO(epot): Validate shape, dtype
+  dca_field = _ArrayFieldMetadata(
+      inner_shape_non_static=shape,
+      dtype=dtype,
+  )
+  return dataclasses.field(**field_kwargs, metadata={_METADATA_KEY: dca_field})
+
+
 class MetaDataclassArray(type):
   """DataclassArray metaclass."""
 
@@ -125,13 +150,22 @@ class MetaDataclassArray(type):
     return Annotated[cls, field_utils.ShapeAnnotation(spec)]
 
 
+# TODO(epot): Restore once pytype support this
+# @typing_extensions.dataclass_transform(
+#     kw_only_default=True,
+#     # TODO(b/272524683):Restore field specifier
+#     # field_specifiers=(
+#     #     dataclasses.Field,
+#     #     dataclasses.field,
+#     #     array_field,
+#     # ),
+# )
 class DataclassArray(metaclass=MetaDataclassArray):
   """Dataclass which behaves like an array.
 
   Usage:
 
   ```python
-  @dataclasses.dataclass
   class Square(DataclassArray):
     pos: f32['*shape 2']
     scale: f32['*shape']
@@ -176,7 +210,6 @@ class DataclassArray(metaclass=MetaDataclassArray):
 
   Field which do not satisfy any of the above conditions are static (including
   field annotated with `field: np.ndarray` or similar).
-
   """
 
   # Child class inherit the default params by default, but can also
@@ -191,12 +224,26 @@ class DataclassArray(metaclass=MetaDataclassArray):
   _shape: Shape
   _xnp: enp.NpModule
 
-  def __init_subclass__(cls, **kwargs):
+  def __init_subclass__(
+      cls,
+      frozen=True,
+      **kwargs,
+  ):
     super().__init_subclass__(**kwargs)
+
+    if not frozen:
+      raise ValueError(f'{cls} cannot be `frozen=False`.')
+
+    # Apply dataclass (in-place)
+    if not typing.TYPE_CHECKING:
+      # TODO(b/227290126): Create pytype issues
+      dataclasses.dataclass(frozen=True)(cls)
+
     # TODO(epot): Could have smart __repr__ which display types if array have
     # too many values (maybe directly in `edc.field(repr=...)`).
-    edc.dataclass(kw_only=True, repr=True)(cls)
-    cls._dca_tree_map_registered = False
+    edc.dataclass(kw_only=True, repr=True, auto_cast=False)(cls)
+    cls._dca_jax_tree_registered = False
+    cls._dca_torch_tree_registered = False
     # Typing annotations have to be lazily evaluated (to support
     # `from __future__ import annotations` and forward reference)
     # To avoid costly `typing.get_type_hints` which perform `eval` and `str`
@@ -208,6 +255,11 @@ class DataclassArray(metaclass=MetaDataclassArray):
     # `__dca_non_init_fields__` (fields should be merged from `.mro()`)
     cls.__dca_non_init_fields__ = set(cls.__dca_non_init_fields__)
 
+  if typing.TYPE_CHECKING:
+    # TODO(b/242839979): pytype do not support PEP 681 -- Data Class Transforms
+    def __init__(self, **kwargs):
+      pass
+
   def __post_init__(self) -> None:
     """Validate and normalize inputs."""
     cls = type(self)
@@ -217,10 +269,20 @@ class DataclassArray(metaclass=MetaDataclassArray):
       _init_cls(self)
 
     # Register the tree_map here instead of `__init_subclass__` as `jax` may
-    # not have been registered yet during import
-    if enp.lazy.has_jax and not cls._dca_tree_map_registered:  # pylint: disable=protected-access
+    # not have been imported yet during import.
+    if enp.lazy.has_jax and not cls._dca_jax_tree_registered:  # pylint: disable=protected-access
       enp.lazy.jax.tree_util.register_pytree_node_class(cls)
-      cls._dca_tree_map_registered = True  # pylint: disable=protected-access
+      cls._dca_jax_tree_registered = True  # pylint: disable=protected-access
+
+    if enp.lazy.has_torch and not cls._dca_torch_tree_registered:  # pylint: disable=protected-access
+      # Note: Torch is updating it's tree API to make it public and use `optree`
+      # as backend: https://github.com/pytorch/pytorch/issues/65761
+      enp.lazy.torch.utils._pytree._register_pytree_node(  # pylint: disable=protected-access
+          cls,
+          flatten_fn=lambda a: a.tree_flatten(),
+          unflatten_fn=lambda vals, ctx: cls.tree_unflatten(ctx, vals),
+      )
+      cls._dca_torch_tree_registered = True  # pylint: disable=protected-access
 
     # Validate and normalize array fields
     # * Maybe cast (list, np) -> xnp
@@ -411,7 +473,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
     }
 
     # Create the new object
-    new_self = dataclasses.replace(self, **init_kwargs)
+    new_self = dataclasses.replace(self, **init_kwargs)  # pytype: disable=wrong-arg-types  # re-none
 
     # TODO(epot): Could try to unify logic bellow with `tree_unflatten`
 
@@ -442,20 +504,38 @@ class DataclassArray(metaclass=MetaDataclassArray):
     """Returns the instance as containing `tf.Tensor`."""
     return self.as_xnp(enp.lazy.tnp)
 
+  def as_torch(self: _DcT) -> _DcT:
+    """Returns the instance as containing `torch.Tensor`."""
+    return self.as_xnp(enp.lazy.torch)
+
   def as_xnp(self: _DcT, xnp: enp.NpModule) -> _DcT:
     """Returns the instance as containing `xnp.ndarray`."""
     if xnp is self.xnp:  # No-op
       return self
+    # Direct `torch` <> `tf`/`jax` conversion not supported, so convert to
+    # `numpy`
+    if enp.lazy.is_torch_xnp(xnp) or enp.lazy.is_torch_xnp(self.xnp):
+
+      def _as_torch(f):
+        arr = np.asarray(f.value)
+        # Torch fail for scalar arrays:
+        # https://github.com/pytorch/pytorch/issues/97021
+        if enp.lazy.is_torch_xnp(xnp) and not arr.shape:  # Destination is torch
+          return xnp.asarray(arr.item(), dtype=lazy.as_torch_dtype(arr.dtype))
+
+        return xnp.asarray(arr)
+
+      array_fn = _as_torch
+    else:
+      array_fn = lambda f: xnp.asarray(f.value)
 
     # Update all childs
     new_self = self._map_field(
-        array_fn=lambda f: xnp.asarray(f.value),
+        array_fn=array_fn,
         dc_fn=lambda f: f.value.as_xnp(xnp),
     )
     return new_self
 
-  # ====== Internal ======
-
   # TODO(pytype): Remove hack. Currently, Python does not support typing
   # annotations for modules, by pytype auto-infer the correct type.
   # So this hack allow auto-completion
@@ -474,6 +554,33 @@ class DataclassArray(metaclass=MetaDataclassArray):
       """Returns the numpy module of the class (np, jnp, tnp)."""
       return self._xnp
 
+  # ====== Torch specific methods ======
+  # Could also add
+  # * x.detach
+  # * x.is_cuda
+  # * x.device
+  # * x.get_device
+
+  def to(self: _DcT, device, **kwargs) -> _DcT:
+    """Move the dataclass array to the device."""
+    if not lazy.is_torch_xnp(self.xnp):
+      raise ValueError('`.to` can only be called when `xnp == torch`')
+    return self.map_field(lambda f: f.to(device, **kwargs))
+
+  def cpu(self: _DcT, *args, **kwargs) -> _DcT:
+    """Move the dataclass array to the CPU device."""
+    if not lazy.is_torch_xnp(self.xnp):
+      raise ValueError('`.cpu` can only be called when `xnp == torch`')
+    return self.map_field(lambda f: f.cpu(*args, **kwargs))
+
+  def cuda(self: _DcT, *args, **kwargs) -> _DcT:
+    """Move the dataclass array to the CUDA device."""
+    if not lazy.is_torch_xnp(self.xnp):
+      raise ValueError('`.cuda` can only be called when `xnp == torch`')
+    return self.map_field(lambda f: f.cuda(*args, **kwargs))
+
+  # ====== Internal ======
+
   @epy.cached_property
   def _all_array_fields(self) -> dict[str, _ArrayField]:
     """All array fields, including `None` values."""
@@ -643,7 +750,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
         array_field_names=list(self._all_array_fields.keys()),
         non_array_field_kwargs={
             f.name: getattr(self, f.name)
-            for f in dataclasses.fields(self)
+            for f in dataclasses.fields(self)  # pytype: disable=wrong-arg-types  # re-none
             if f.name not in self._all_array_fields  # pylint: disable=unsupported-membership-test
         },
     )
@@ -664,7 +771,7 @@ class DataclassArray(metaclass=MetaDataclassArray):
     )
     init_fields = {}
     non_init_fields = {}
-    fields = {f.name: f for f in dataclasses.fields(cls)}
+    fields = {f.name: f for f in dataclasses.fields(cls)}  # pytype: disable=wrong-arg-types  # re-none
     for k, v in metadata.non_array_field_kwargs.items():
       if fields[k].init:
         init_fields[k] = v
@@ -704,7 +811,6 @@ def _init_cls(self: DataclassArray) -> None:
 
   This will:
 
-  * Validate the `@dataclass(frozen=True)` is correctly applied
   * Extract the types annotations, detect which fields are arrays or static,
     and store the result in `_dca_fields_metadata`
   * For static `DataclassArray` (class with only static fields), it will
@@ -716,12 +822,6 @@ def _init_cls(self: DataclassArray) -> None:
   """
   cls = type(self)
 
-  # Make sure the dataclass was registered and frozen
-  if not dataclasses.is_dataclass(cls) or not cls.__dataclass_params__.frozen:  # pytype: disable=attribute-error
-    raise ValueError(
-        '`dca.DataclassArray` need to be @dataclasses.dataclass(frozen=True)'
-    )
-
   # The first time, compute typing annotations & metadata
   # At this point, `ForwardRef` should have been resolved.
   try:
@@ -746,7 +846,7 @@ def _init_cls(self: DataclassArray) -> None:
     )
 
   dca_fields_metadata = {
-      f.name: _make_field_metadata(f, hints) for f in dataclasses.fields(cls)
+      f.name: _make_field_metadata(f, hints) for f in dataclasses.fields(cls)  # pytype: disable=wrong-arg-types
   }
   dca_fields_metadata = {  # Filter `None` values (static fields)
       k: v for k, v in dca_fields_metadata.items() if v is not None
@@ -858,31 +958,6 @@ class _TreeMetadata:
   non_array_field_kwargs: dict[str, Any]
 
 
-def array_field(
-    shape: Shape,
-    dtype: DTypeArg = float,
-    **field_kwargs,
-) -> dataclasses.Field[DcOrArray]:
-  """Dataclass array field.
-
-  See `dca.DataclassArray` for example.
-
-  Args:
-    shape: Inner shape of the field
-    dtype: Type of the field
-    **field_kwargs: Args forwarded to `dataclasses.field`
-
-  Returns:
-    The dataclass field.
-  """
-  # TODO(epot): Validate shape, dtype
-  dca_field = _ArrayFieldMetadata(
-      inner_shape_non_static=shape,
-      dtype=dtype,
-  )
-  return dataclasses.field(**field_kwargs, metadata={_METADATA_KEY: dca_field})
-
-
 # TODO(epot): Should refactor `_ArrayField` in `_DataclassArrayField` and
 # `_ArrayField` depending on whether dtype is `DataclassArray` or not.
 # Alternativelly, maybe should create a `DcArrayDType` dtype instead.
